@@ -1,3 +1,16 @@
+/*
+ * LibHTP (http://www.libhtp.org)
+ * Copyright 2009,2010 Ivan Ristic <ivanr@webkreator.com>
+ *
+ * LibHTP is an open source product, released under terms of the General Public Licence
+ * version 2 (GPLv2). Please refer to the file LICENSE, which contains the complete text
+ * of the license.
+ *
+ * In addition, there is a special exception that allows LibHTP to be freely
+ * used with any OSI-approved open source licence. Please refer to the file
+ * LIBHTP_LICENSING_EXCEPTION for the full text of the exception.
+ *
+ */
 
 #include "htp.h"
 
@@ -145,9 +158,11 @@ htp_cfg_t *htp_config_create() {
  * @return A copy of the configuration structure.
  */
 htp_cfg_t *htp_config_copy(htp_cfg_t *cfg) {
-    htp_cfg_t *copy = calloc(1, sizeof(htp_cfg_t));
+    htp_cfg_t *copy = malloc(sizeof(htp_cfg_t));
     if (copy == NULL) return NULL;
 
+    *copy = *cfg;
+
     // Create copies of the hooks' structures
     if (cfg->hook_transaction_start != NULL) {
         copy->hook_transaction_start = hook_copy(cfg->hook_transaction_start);
@@ -165,6 +180,14 @@ htp_cfg_t *htp_config_copy(htp_cfg_t *cfg) {
         }
     }
 
+    if (cfg->hook_request_uri_normalize != NULL) {
+        copy->hook_request_uri_normalize = hook_copy(cfg->hook_request_uri_normalize);
+        if (copy->hook_request_uri_normalize == NULL) {
+            free(copy);
+            return NULL;
+        }
+    }
+
     if (cfg->hook_request_headers != NULL) {
         copy->hook_request_headers = hook_copy(cfg->hook_request_headers);
         if (copy->hook_request_headers == NULL) {
@@ -257,6 +280,7 @@ void htp_config_destroy(htp_cfg_t *cfg) {
     // Destroy the hooks
     hook_destroy(cfg->hook_transaction_start);
     hook_destroy(cfg->hook_request_line);
+    hook_destroy(cfg->hook_request_uri_normalize);
     hook_destroy(cfg->hook_request_headers);
     hook_destroy(cfg->hook_request_body_data);
     hook_destroy(cfg->hook_request_trailer);
@@ -292,6 +316,16 @@ void htp_config_register_request_line(htp_cfg_t *cfg, int (*callback_fn)(htp_con
     hook_register(&cfg->hook_request_line, callback_fn);
 }
 
+/**
+ * Registers a request_uri_normalize callback.
+ *
+ * @param cfg
+ * @param callback_fn 
+ */
+void htp_config_register_request_uri_normalize(htp_cfg_t *cfg, int (*callback_fn)(htp_connp_t *)) {
+    hook_register(&cfg->hook_request_uri_normalize, callback_fn);
+}
+
 /**
  * Registers a request_headers callback.
  *
@@ -406,6 +440,16 @@ void htp_config_set_bestfit_map(htp_cfg_t *cfg, unsigned char *map) {
     cfg->path_u_bestfit_map = map;
 }
 
+/**
+ * Whether to generate the request_uri_normalized field.
+ *
+ * @param cfg
+ * @param generate
+ */
+void htp_config_set_generate_request_uri_normalized(htp_cfg_t *cfg, int generate) {
+    cfg->generate_request_uri_normalized = generate;
+}
+
 /**
  * Configures whether backslash characters are treated as path segment separators. They
  * are not on Unix systems, but are on Windows systems. If this setting is enabled, a path
@@ -418,6 +462,18 @@ void htp_config_set_path_backslash_separators(htp_cfg_t *cfg, int backslash_sepa
     cfg->path_backslash_separators = backslash_separators;
 }
 
+/**
+ * Configures whether backslash characters are treated as query segment separators. They
+ * are not on Unix systems, but are on Windows systems. If this setting is enabled, a query
+ * such as "/one\two/three" will be converted to "/one/two/three".
+ *
+ * @param cfg
+ * @param backslash_separators
+ */
+void htp_config_set_query_backslash_separators(htp_cfg_t *cfg, int backslash_separators) {
+    cfg->query_backslash_separators = backslash_separators;
+}
+
 /**
  * Configures filesystem sensitivity. This setting affects
  * how URL paths are normalized. There are no path modifications by default, but
@@ -430,6 +486,18 @@ void htp_config_set_path_case_insensitive(htp_cfg_t *cfg, int case_insensitive)
     cfg->path_case_insensitive = case_insensitive;
 }
 
+/**
+ * Configures filesystem sensitivity. This setting affects
+ * how URL querys are normalized. There are no query modifications by default, but
+ * on a case-insensitive systems query will be converted to lowercase.
+ *
+ * @param cfg
+ * @param case_insensitive
+ */
+void htp_config_set_query_case_insensitive(htp_cfg_t *cfg, int case_insensitive) {
+    cfg->query_case_insensitive = case_insensitive;
+}
+
 /**
  * Configures whether consecutive path segment separators will be compressed. When
  * enabled, a path such as "/one//two" will be normalized to "/one/two". The backslash_separators
@@ -444,6 +512,20 @@ void htp_config_set_path_compress_separators(htp_cfg_t *cfg, int compress_separa
     cfg->path_compress_separators = compress_separators;
 }
 
+/**
+ * Configures whether consecutive query segment separators will be compressed. When
+ * enabled, a query such as "/one//two" will be normalized to "/one/two". The backslash_separators
+ * and decode_separators parameters are used before compression takes place. For example, if
+ * backshasl_deparators and decode_separators are both enabled, the query "/one\\/two\/%5cthree/%2f//four"
+ * will be converted to "/one/two/three/four".
+ *
+ * @param cfg
+ * @param compress_separators
+ */
+void htp_config_set_query_compress_separators(htp_cfg_t *cfg, int compress_separators) {
+    cfg->query_compress_separators = compress_separators;
+}
+
 /**
  * This parameter is used to predict how a server will react when control
  * characters are present in a request path, but does not affect path
@@ -458,6 +540,20 @@ void htp_config_set_path_control_char_handling(htp_cfg_t *cfg, int control_char_
     cfg->path_control_char_handling = control_char_handling;
 }
 
+/**
+ * This parameter is used to predict how a server will react when control
+ * characters are present in a request query, but does not affect query
+ * normalization.
+ *
+ * @param cfg
+ * @param control_char_handling Use NONE with servers that ignore control characters in
+ *                              request query, and STATUS_400 with servers that respond
+ *                              with 400.
+ */
+void htp_config_set_query_control_char_handling(htp_cfg_t *cfg, int control_char_handling) {
+    cfg->query_control_char_handling = control_char_handling;
+}
+
 /**
  * Controls the UTF-8 treatment of request paths. One option is to only validate
  * path as UTF-8. In this case, the UTF-8 flags will be raised as appropriate, and
@@ -484,6 +580,19 @@ void htp_config_set_path_decode_separators(htp_cfg_t *cfg, int decode_separators
     cfg->path_decode_separators = decode_separators;
 }
 
+/**
+ * Configures whether encoded query segment separators will be decoded. Apache does
+ * not do this, but IIS does. If enabled, a query such as "/one%2ftwo" will be normalized
+ * to "/one/two". If the backslash_separators option is also enabled, encoded backslash
+ * characters will be converted too (and subseqently normalized to forward slashes).
+ *
+ * @param cfg
+ * @param decode_separators
+ */
+void htp_config_set_query_decode_separators(htp_cfg_t *cfg, int decode_separators) {
+    cfg->query_decode_separators = decode_separators;
+}
+
 /**
  * Configures whether %u-encoded sequences in path will be decoded. Such sequences
  * will be treated as invalid URL encoding if decoding is not desireable. 
@@ -495,6 +604,17 @@ void htp_config_set_path_decode_u_encoding(htp_cfg_t *cfg, int decode_u_encoding
     cfg->path_decode_u_encoding = decode_u_encoding;
 }
 
+/**
+ * Configures whether %u-encoded sequences in query will be decoded. Such sequences
+ * will be treated as invalid URL encoding if decoding is not desireable.
+ *
+ * @param cfg
+ * @param decode_u_encoding
+ */
+void htp_config_set_query_decode_u_encoding(htp_cfg_t *cfg, int decode_u_encoding) {
+    cfg->query_decode_u_encoding = decode_u_encoding;
+}
+
 /**
  * Configures how server reacts to invalid encoding in path.
  *
@@ -507,6 +627,17 @@ void htp_config_set_path_invalid_encoding_handling(htp_cfg_t *cfg, int invalid_e
     cfg->path_invalid_encoding_handling = invalid_encoding_handling;
 }
 
+/**
+ * Configures how server reacts to invalid encoding in query.
+ *
+ * @param cfg
+ * @param invalid_encoding_handling The available options are: URL_DECODER_PRESERVE_PERCENT,
+ *                                  URL_DECODER_REMOVE_PERCENT, URL_DECODER_DECODE_INVALID
+ *                                  and URL_DECODER_STATUS_400.
+ */
+void htp_config_set_query_invalid_encoding_handling(htp_cfg_t *cfg, int invalid_encoding_handling) {
+    cfg->query_invalid_encoding_handling = invalid_encoding_handling;
+}
 
 
 /**
@@ -533,6 +664,18 @@ void htp_config_set_path_nul_encoded_handling(htp_cfg_t *cfg, int nul_encoded_ha
     cfg->path_nul_encoded_handling = nul_encoded_handling;
 }
 
+/**
+ * Configures how server reacts to encoded NUL bytes. Some servers will terminate
+ * query at NUL, while some will respond with 400 or 404. When the termination option
+ * is not used, the NUL byte will remain in the query.
+ *
+ * @param cfg
+ * @param nul_encoded_handling Possible values: TERMINATE, STATUS_400, STATUS_404
+ */
+void htp_config_set_query_nul_encoded_handling(htp_cfg_t *cfg, int nul_encoded_handling) {
+    cfg->query_nul_encoded_handling = nul_encoded_handling;
+}
+
 /**
  * Configures how server reacts to raw NUL bytes. Some servers will terminate
  * path at NUL, while some will respond with 400 or 404. When the termination option
@@ -545,6 +688,18 @@ void htp_config_set_path_nul_raw_handling(htp_cfg_t *cfg, int nul_raw_handling)
     cfg->path_nul_raw_handling = nul_raw_handling;
 }
 
+/**
+ * Configures how server reacts to raw NUL bytes. Some servers will terminate
+ * query at NUL, while some will respond with 400 or 404. When the termination option
+ * is not used, the NUL byte will remain in the query.
+ *
+ * @param cfg
+ * @param nul_raw_handling Possible values: TERMINATE, STATUS_400, STATUS_404
+ */
+void htp_config_set_query_nul_raw_handling(htp_cfg_t *cfg, int nul_raw_handling) {
+    cfg->query_nul_raw_handling = nul_raw_handling;
+}
+
 /**
  * Sets the replacement characater that will be used to in the lossy best-fit
  * mapping from Unicode characters into single-byte streams. The question mark
@@ -607,6 +762,10 @@ int htp_config_set_server_personality(htp_cfg_t *cfg, int personality) {
             cfg->path_backslash_separators = YES;
             cfg->path_decode_separators = YES;
             cfg->path_compress_separators = YES;
+
+//            cfg->query_backslash_separators = YES;
+            cfg->query_decode_separators = YES;
+//            cfg->query_compress_separators = YES;
             break;
 
         case HTP_SERVER_IDS:
@@ -622,6 +781,12 @@ int htp_config_set_server_personality(htp_cfg_t *cfg, int personality) {
             cfg->path_decode_u_encoding = YES;
             cfg->path_unicode_mapping = BESTFIT;
             cfg->path_convert_utf8 = YES;
+
+//            cfg->query_backslash_separators = YES;
+            cfg->query_case_insensitive = YES;
+            cfg->query_decode_separators = YES;
+//            cfg->query_compress_separators = YES;
+            cfg->query_decode_u_encoding = YES;
             break;
 
         case HTP_SERVER_APACHE :
@@ -636,6 +801,12 @@ int htp_config_set_server_personality(htp_cfg_t *cfg, int personality) {
             cfg->path_compress_separators = YES;
             cfg->path_invalid_encoding_handling = URL_DECODER_STATUS_400;            
             cfg->path_control_char_handling = NONE;
+
+//            cfg->query_backslash_separators = NO;
+            cfg->query_decode_separators = NO;
+//            cfg->query_compress_separators = YES;
+            cfg->query_invalid_encoding_handling = URL_DECODER_STATUS_400;
+            cfg->query_control_char_handling = NONE;
             break;
 
         case HTP_SERVER_IIS_5_1:
@@ -651,6 +822,14 @@ int htp_config_set_server_personality(htp_cfg_t *cfg, int personality) {
             cfg->path_decode_u_encoding = YES;
             cfg->path_unicode_mapping = BESTFIT;
             cfg->path_control_char_handling = NONE;
+
+//            cfg->query_backslash_separators = YES;
+            cfg->query_decode_separators = NO;
+//            cfg->query_compress_separators = YES;
+            cfg->query_invalid_encoding_handling = URL_DECODER_PRESERVE_PERCENT;
+            cfg->query_decode_u_encoding = YES;
+//            cfg->query_unicode_mapping = BESTFIT;
+            cfg->query_control_char_handling = NONE;
             break;
 
         case HTP_SERVER_IIS_6_0:
@@ -666,6 +845,14 @@ int htp_config_set_server_personality(htp_cfg_t *cfg, int personality) {
             cfg->path_decode_u_encoding = YES;
             cfg->path_unicode_mapping = STATUS_400;
             cfg->path_control_char_handling = STATUS_400;
+
+//            cfg->query_backslash_separators = YES;
+            cfg->query_decode_separators = YES;
+//            cfg->query_compress_separators = YES;
+            cfg->query_invalid_encoding_handling = URL_DECODER_STATUS_400;
+            cfg->query_decode_u_encoding = YES;
+//            cfg->query_unicode_mapping = STATUS_400;
+            cfg->query_control_char_handling = STATUS_400;
             break;
 
         case HTP_SERVER_IIS_7_0:
@@ -680,6 +867,12 @@ int htp_config_set_server_personality(htp_cfg_t *cfg, int personality) {
             cfg->path_compress_separators = YES;
             cfg->path_invalid_encoding_handling = URL_DECODER_STATUS_400;
             cfg->path_control_char_handling = STATUS_400;
+
+//            cfg->query_backslash_separators = YES;
+            cfg->query_decode_separators = YES;
+//            cfg->query_compress_separators = YES;
+            cfg->query_invalid_encoding_handling = URL_DECODER_STATUS_400;
+            cfg->query_control_char_handling = STATUS_400;
             break;
             
         default:
