@@ -1,8 +1,84 @@
+/*
+ * LibHTP (http://www.libhtp.org)
+ * Copyright 2009,2010 Ivan Ristic <ivanr@webkreator.com>
+ *
+ * LibHTP is an open source product, released under terms of the General Public Licence
+ * version 2 (GPLv2). Please refer to the file LICENSE, which contains the complete text
+ * of the license.
+ *
+ * In addition, there is a special exception that allows LibHTP to be freely
+ * used with any OSI-approved open source licence. Please refer to the file
+ * LIBHTP_LICENSING_EXCEPTION for the full text of the exception.
+ *
+ */
 
 #include <stdlib.h>
 
 #include "htp.h"
 
+/**
+ * Performs check for a CONNECT transaction to decide whether inbound
+ * parsing needs to be suspended.
+ *
+ * @param connp
+ * @return HTP_OK if the request does not use CONNECT, HTP_DATA_OTHER if
+ *          inbound parsing needs to be suspended until we hear from the
+ *          other side
+ */
+int htp_connp_REQ_CONNECT_CHECK(htp_connp_t *connp) {
+    // If the request uses the CONNECT method, then there will
+    // not be a request body, but first we need to wait to see the
+    // response in order to determine if the tunneling request
+    // was a success.
+    if (connp->in_tx->request_method_number == M_CONNECT) {
+        connp->in_state = htp_connp_REQ_CONNECT_WAIT_RESPONSE;
+        connp->in_status = STREAM_STATE_DATA_OTHER;
+        connp->in_tx->progress = TX_PROGRESS_WAIT;
+
+        return HTP_DATA_OTHER;
+    }
+
+    // Continue to the next step to determine the presence
+    // of the request body
+    connp->in_state = htp_connp_REQ_BODY_DETERMINE;
+
+    return HTP_OK;
+}
+
+/**
+ * Determines whether inbound parsing, which was suspended after
+ * encountering a CONNECT transaction, can proceed (after receiving
+ * the response).
+ *
+ * @param connp
+ * @return HTP_OK if the parser can resume parsing, HTP_DATA_OTHER if
+ *         it needs to continue waiting.
+ */
+int htp_connp_REQ_CONNECT_WAIT_RESPONSE(htp_connp_t *connp) {
+    // Check that we saw the response line of the current
+    // inbound transaction.
+    if (connp->in_tx->progress <= TX_PROGRESS_RES_LINE) {
+        return HTP_DATA_OTHER;
+    }
+
+    // A 2xx response means a tunnel was established. Anything
+    // else means we continue to follow the HTTP stream.
+    if ((connp->in_tx->response_status_number >= 200) && (connp->in_tx->response_status_number <= 299)) {
+        // TODO Check that the server did not accept a connection
+        //      to itself.
+
+        // The requested tunnel was established: we are going
+        // to ignore the remaining data on this stream
+        connp->in_status = STREAM_STATE_TUNNEL;
+        connp->in_state = htp_connp_REQ_IDLE;
+    } else {
+        // No tunnel; continue to the next transaction
+        connp->in_state = htp_connp_REQ_IDLE;
+    }
+
+    return HTP_OK;
+}
+
 /**
  * Consumes bytes until the end of the current line.
  *
@@ -188,9 +264,9 @@ int htp_connp_REQ_BODY_DETERMINE(htp_connp_t *connp) {
     htp_header_t *cl = table_getc(connp->in_tx->request_headers, "content-length");
     htp_header_t *te = table_getc(connp->in_tx->request_headers, "transfer-encoding");
 
-    // First check for the Transfer-Encoding header, which
+    // Check for the Transfer-Encoding header, which
     // would indicate a chunked request body
-    if (te != NULL) {
+    if (te != NULL && te->value != NULL) {
         // Make sure it contains "chunked" only
         if (bstr_cmpc(te->value, "chunked") != 0) {
             // Invalid T-E header value
@@ -223,7 +299,7 @@ int htp_connp_REQ_BODY_DETERMINE(htp_connp_t *connp) {
         connp->in_tx->progress = TX_PROGRESS_REQ_BODY;
     } else
         // Next check for the presence of the Content-Length header
-        if (cl != NULL) {
+        if (cl != NULL && cl->value != NULL) {
         // It seems that we have a request body.
         connp->in_tx->request_transfer_coding = IDENTITY;
 
@@ -282,8 +358,8 @@ int htp_connp_REQ_BODY_DETERMINE(htp_connp_t *connp) {
             // There is no host information in the URI. Place the
             // hostname from the headers into the parsed_uri structure.
             htp_replace_hostname(connp, connp->in_tx->parsed_uri, h->value);
-        } else {
-            // The host information is present both in the
+        } else if (bstr_cmp_nocase(h->value, connp->in_tx->parsed_uri->hostname) != 0) {
+            // The host information is different in the
             // headers and the URI. The HTTP RFC states that
             // we should ignore the headers copy.
             connp->in_tx->flags |= HTP_AMBIGUOUS_HOST;
@@ -332,9 +408,9 @@ int htp_connp_REQ_HEADERS(htp_connp_t *connp) {
 
         // Have we reached the end of the line?
         if (connp->in_next_byte == LF) {
-#ifdef HTP_DEBUG
+            #ifdef HTP_DEBUG
             fprint_raw_data(stderr, __FUNCTION__, connp->in_line, connp->in_line_len);
-#endif
+            #endif
 
             // Should we terminate headers?
             if (htp_connp_is_line_terminator(connp, connp->in_line, connp->in_line_len)) {
@@ -354,9 +430,7 @@ int htp_connp_REQ_HEADERS(htp_connp_t *connp) {
                 // Cleanup
                 free(connp->in_header_line);
                 connp->in_line_len = 0;
-                connp->in_header_line = NULL;
-                connp->in_header_line_index = -1;
-                connp->in_header_line_counter = 0;
+                connp->in_header_line = NULL;                
 
                 // We've seen all request headers
                 if (connp->in_chunk_count != connp->in_chunk_request_index) {
@@ -366,7 +440,8 @@ int htp_connp_REQ_HEADERS(htp_connp_t *connp) {
                 // Move onto the next processing phase
                 if (connp->in_tx->progress == TX_PROGRESS_REQ_HEADERS) {
                     // Determine if this request has a body
-                    connp->in_state = htp_connp_REQ_BODY_DETERMINE;
+                    //connp->in_state = htp_connp_REQ_BODY_DETERMINE;
+                    connp->in_state = htp_connp_REQ_CONNECT_CHECK;
                 } else {
                     // Run hook REQUEST_TRAILER
                     int rc = hook_run_all(connp->cfg->hook_request_trailer, connp);
@@ -385,6 +460,7 @@ int htp_connp_REQ_HEADERS(htp_connp_t *connp) {
             }
 
             // Prepare line for consumption
+            size_t raw_in_line_len = connp->in_line_len;
             htp_chomp(connp->in_line, &connp->in_line_len);
 
             // Check for header folding
@@ -416,7 +492,27 @@ int htp_connp_REQ_HEADERS(htp_connp_t *connp) {
             }
 
             // Add the raw header line to the list
+            if (raw_in_line_len > connp->in_line_len) {
+                if (raw_in_line_len - connp->in_line_len == 2 &&
+                        connp->in_line[connp->in_line_len] == 0x0d &&
+                        connp->in_line[connp->in_line_len + 1] == 0x0a) {
+                    connp->in_header_line->terminators = NULL;
+                } else {
+                    connp->in_header_line->terminators =
+                        bstr_memdup((char *) connp->in_line + connp->in_line_len,
+                                raw_in_line_len - connp->in_line_len);
+                    if (connp->in_header_line->terminators == NULL) {
+                        return HTP_ERROR;
+                    }
+                }
+            } else {
+                connp->in_header_line->terminators = NULL;
+            }
+
             connp->in_header_line->line = bstr_memdup((char *) connp->in_line, connp->in_line_len);
+            if (connp->in_header_line->line == NULL) {
+                return HTP_ERROR;
+            }
             list_add(connp->in_tx->request_header_lines, connp->in_header_line);
             connp->in_header_line = NULL;
 
@@ -477,9 +573,9 @@ int htp_connp_REQ_LINE(htp_connp_t *connp) {
 
         // Have we reached the end of the line?
         if (connp->in_next_byte == LF) {
-#ifdef HTP_DEBUG
+            #ifdef HTP_DEBUG
             fprint_raw_data(stderr, __FUNCTION__, connp->in_line, connp->in_line_len);
-#endif
+            #endif
 
             // Is this a line that should be ignored?
             if (htp_connp_is_line_ignorable(connp, connp->in_line, connp->in_line_len)) {
@@ -498,6 +594,9 @@ int htp_connp_REQ_LINE(htp_connp_t *connp) {
 
             htp_chomp(connp->in_line, &connp->in_line_len);
             connp->in_tx->request_line = bstr_memdup((char *) connp->in_line, connp->in_line_len);
+            if (connp->in_tx->request_line == NULL) {
+                return HTP_ERROR;
+            }
 
             // Parse request line
             if (connp->cfg->parse_request_line(connp) != HTP_OK) {
@@ -516,15 +615,41 @@ int htp_connp_REQ_LINE(htp_connp_t *connp) {
                 if (htp_parse_uri(connp->in_tx->request_uri, &(connp->in_tx->parsed_uri_incomplete)) != HTP_OK) {
                     // Note: downstream responsible for error logging
                     return HTP_ERROR;
-                }
+                }               
 
                 // Keep the original URI components, but
                 // create a copy which we can normalize and use internally
                 if (htp_normalize_parsed_uri(connp, connp->in_tx->parsed_uri_incomplete, connp->in_tx->parsed_uri)) {
                     // Note: downstream responsible for error logging
                     return HTP_ERROR;
+                }               
+
+                // Run hook REQUEST_URI_NORMALIZE
+                int rc = hook_run_all(connp->cfg->hook_request_uri_normalize, connp);
+                if (rc != HOOK_OK) {
+                    htp_log(connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0,
+                            "Request URI normalize callback returned error (%d)", rc);
+                    return HTP_ERROR;
                 }
 
+                // Now is a good time to generate request_uri_normalized, before we finalize
+                // parsed_uri (and lose the information which parts were provided in the request and
+                // which parts we added).
+                if (connp->cfg->generate_request_uri_normalized) {                 
+                    connp->in_tx->request_uri_normalized = htp_unparse_uri_noencode(connp->in_tx->parsed_uri);                   
+
+                    if (connp->in_tx->request_uri_normalized == NULL) {
+                        // There's no sense in logging anything on a memory allocation failure
+                        return HTP_ERROR;
+                    }                   
+
+                    #ifdef HTP_DEBUG
+                    fprint_raw_data(stderr, "request_uri_normalized",
+                        (unsigned char *) bstr_ptr(connp->in_tx->request_uri_normalized),
+                        bstr_len(connp->in_tx->request_uri_normalized));
+                    #endif
+                }               
+
                 // Finalize parsed_uri
 
                 // Scheme
@@ -534,6 +659,9 @@ int htp_connp_REQ_LINE(htp_connp_t *connp) {
                     }
                 } else {
                     connp->in_tx->parsed_uri->scheme = bstr_cstrdup("http");
+                    if (connp->in_tx->parsed_uri->scheme == NULL) {
+                        return HTP_ERROR;
+                    }
                 }
 
                 // Port
@@ -559,7 +687,10 @@ int htp_connp_REQ_LINE(htp_connp_t *connp) {
                 // Path
                 if (connp->in_tx->parsed_uri->path == NULL) {
                     connp->in_tx->parsed_uri->path = bstr_cstrdup("/");
-                }
+                    if (connp->in_tx->parsed_uri->path == NULL) {
+                        return HTP_ERROR;
+                    }
+                }               
             }
 
             // Run hook REQUEST_LINE
@@ -647,6 +778,10 @@ int htp_connp_REQ_IDLE(htp_connp_t * connp) {
     return HTP_OK;
 }
 
+size_t htp_connp_req_data_consumed(htp_connp_t *connp) {
+    return connp->in_current_offset;
+}
+
 /**
  * Process a chunk of inbound (client or request) data.
  * 
@@ -657,17 +792,34 @@ int htp_connp_REQ_IDLE(htp_connp_t * connp) {
  * @return HTP_OK on state change, HTTP_ERROR on error, or HTP_DATA when more data is needed.
  */
 int htp_connp_req_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *data, size_t len) {
-#ifdef HTP_DEBUG
+    #ifdef HTP_DEBUG
     fprintf(stderr, "htp_connp_req_data(connp->in_status %x)\n", connp->in_status);
     fprint_raw_data(stderr, __FUNCTION__, data, len);
-#endif    
+    #endif
 
     // Return if the connection has had a fatal error
-    if (connp->in_status != STREAM_STATE_OPEN) {
-        // We allow calls that allow the parser to finalize their work
-        if (!((connp->out_status == STREAM_STATE_CLOSED) && (len == 0))) {
-            return STREAM_STATE_ERROR;
-        }
+    if (connp->in_status == STREAM_STATE_ERROR) {
+        htp_log(connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0, "Inbound parser is in STREAM_STATE_ERROR");
+
+        #ifdef HTP_DEBUG
+        fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_DATA (previous error)\n");
+        #endif
+
+        return STREAM_STATE_ERROR;
+    }
+
+    // If the length of the supplied data chunk is zero, proceed
+    // only if the stream has been closed. We do not allow zero-sized
+    // chunks in the API, but we use it internally to force the parsers
+    // to finalize parsing.
+    if ((len == 0) && (connp->in_status != STREAM_STATE_CLOSED)) {
+        htp_log(connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0, "Zero-length data chunks are not allowed");
+
+        #ifdef HTP_DEBUG
+        fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_DATA (zero-length chunk)\n");
+        #endif
+
+        return STREAM_STATE_ERROR;
     }
 
     // Store the current chunk information
@@ -679,32 +831,76 @@ int htp_connp_req_data(htp_connp_t *connp, htp_time_t timestamp, unsigned char *
     connp->conn->in_data_counter += len;
     connp->conn->in_packet_counter++;
 
+    // Return without processing any data if the stream is in tunneling
+    // mode (which it would be after an initial CONNECT transaction).
+    if (connp->in_status == STREAM_STATE_TUNNEL) {
+        #ifdef HTP_DEBUG
+        fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_TUNNEL\n");
+        #endif
+        return STREAM_STATE_TUNNEL;
+    }
+
     // Invoke a processor, in a loop, until an error
     // occurs or until we run out of data. Many processors
     // will process a request, each pointing to the next
     // processor that needs to run.
     for (;;) {
-#ifdef HTP_DEBUG
+        #ifdef HTP_DEBUG
         fprintf(stderr, "htp_connp_req_data: in state=%s, progress=%s\n",
             htp_connp_in_state_as_string(connp),
             htp_tx_progress_as_string(connp->in_tx));
-#endif
+        #endif
 
         // Return if there's been an error
         // or if we've run out of data. We are relying
         // on processors to add error messages, so we'll
         // keep quiet here.
         int rc = connp->in_state(connp);
-        if (rc != HTP_OK) {
+        if (rc == HTP_OK) {
+            if (connp->in_status == STREAM_STATE_TUNNEL) {
+                #ifdef HTP_DEBUG
+                fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_TUNNEL\n");
+                #endif
+
+                return STREAM_STATE_TUNNEL;
+            }
+        } else {
             // Do we need more data?
             if (rc == HTP_DATA) {
+                #ifdef HTP_DEBUG
+                fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_DATA\n");
+                #endif
+
                 return STREAM_STATE_DATA;
             }
 
+            // Check for suspended parsing
+            if (rc == HTP_DATA_OTHER) {
+                // We might have actually consumed the entire data chunk?
+                if (connp->in_current_offset >= connp->in_current_len) {
+                    // Do not send STREAM_DATE_DATA_OTHER if we've
+                    // consumed the entire chunk
+                    #ifdef HTP_DEBUG
+                    fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_DATA (suspended parsing)\n");
+                    #endif
+                    return STREAM_STATE_DATA;
+                } else {
+                    // Partial chunk consumption
+                    #ifdef HTP_DEBUG
+                    fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_DATA_OTHER\n");
+                    #endif
+                    return STREAM_STATE_DATA_OTHER;
+                }
+            }
+
             // Remember that we've had an error. Errors are
-            // not possible to recover from.
+            // (at least at present) not possible to recover from.
             connp->in_status = STREAM_STATE_ERROR;
 
+            #ifdef HTP_DEBUG
+            fprintf(stderr, "htp_connp_req_data: returning STREAM_STATE_ERROR (state response)\n");
+            #endif
+
             return STREAM_STATE_ERROR;
         }
     }
